package org.artifactory.projects;

import lombok.NonNull;
import org.artifactory.common.MutableStatusHolder;
import org.artifactory.repo.RepoPath;
import org.artifactory.security.SimpleUser;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.spring.ReloadableBean;
import org.jfrog.access.client.AccessClientException;
import org.jfrog.access.client.project.ProjectEntity;
import org.jfrog.access.client.project.ProjectResourceEntity;
import org.jfrog.access.client.project.ProjectSlimEntity;
import org.jfrog.access.common.EnvironmentsConstants;
import org.jfrog.access.common.ResourceType;
import org.jfrog.access.rest.user.User;

import javax.annotation.Nullable;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author BarakH
 */
public interface ProjectsService extends ReloadableBean, ContextCreationListener {

    String PROJECT = "project";
    String PROJECT_KEY = "projectKey";
    String DEFAULT_PROJECT = "default";
    String UNASSIGNED = "unassigned";
    String ENVIRONMENTS = "environments";
    String ENV_DEV = EnvironmentsConstants.DEV;
    String ENV_PROD = EnvironmentsConstants.PROD;
    String DEFAULT_ENVIRONMENT = ENV_DEV;
    Set<String> GLOBAL_ENVIRONMENTS = Set.of(ENV_PROD, ENV_DEV);
    long UNLIMITED_STORAGE = -1L;

    default void syncProjectResources(ResourceType resourceType, Set<String> resources) {
    }

    /**
     * Add a new repository to the given projectKey and environment
     */
    default void addProjectRepository(String repoKey, String projectKey, Set<String> environments) throws
            AccessClientException {
    }

    /**
     * Update the current repository environment in the projectKey
     */
    default void updateProjectRepository(String repoKey, Set<String> environments) throws
            AccessClientException {
    }

    /**
     * Remove a repo resource from its current projectKey
     */
    default void removeRepositoryFromCurrentProject(String repoKey) throws AccessClientException {
    }

    /**
     * @return A list with all the projects keys the user has ProjectAdmin role on (including
     * {@link #DEFAULT_PROJECT} for super admin).
     */
    @NonNull
    List<String> getAllProjectsManagedByUser();

    /**
     * @return A list with all the projects keys
     */
    @NonNull
    List<String> getAllProjects();

    /**
     * @return A list of all the projects models from the DB.
     * The ProjectEntity contains all the projectKey configuration including the members, resources and roles.
     */
    @NonNull
    List<ProjectEntity> getAllProjectsEntities();

    /**
     * @return A set of all the projectKey related repositories keys (Assigned and shared with)
     */
    @NonNull
    Set<String> getProjectRepositoriesKeys(String projectKey, boolean ignoreCache);

    /**
     * @return The repository projectKey or {@link #DEFAULT_PROJECT} if none
     */
    @NonNull
    String getProjectKeyOfRepository(String repoKey);

    /**
     * @return The repository projectKey or {@link #DEFAULT_PROJECT} if none. Skips the license validation
     */
    @NonNull
    String getProjectKeyOfRepositoryInternal(@Nullable String repoKey);

    /**
     * @return The projectName or {@link #DEFAULT_PROJECT} if none
     */
    @NonNull
    String getProjectDisplayName(String projectKey);

    /**
     * @return Map of <repoKey , projectKey>
     */
    @NonNull
    Map<String, String> getRepoKeyToProjectMap(@NonNull Set<String> repositories, boolean ignoreCache);

    /**
     * Each repository can be assigned to one projectKey and meanwhile shared with many others.
     * The ProjectResourceEntity holds all the relevant information about the repository state withing the projectKey.
     *
     * @return The ProjectResourceEntity of the given repository within the given projectKey or null.
     */
    @Nullable
    ProjectResourceEntity getProjectRepositoryEntity(@NonNull String projectKey, @NonNull String repoKey);

    /**
     * @return Project general configuration
     */
    @Nullable
    ProjectSlimEntity getProjectSpec(String projectKey);

    /**
     * @return Map of <projectKey , <repoKey , ProjectResourceEntity>>
     */
    @NonNull
    Map<String, Map<String, ProjectResourceEntity>> getAllProjectsRepositoriesMap(boolean ignoreCache);

    /**
     * @return Map of <repoKey , ProjectResourceEntity>
     */
    @NonNull
    Map<String, ProjectResourceEntity> getProjectRepositoriesMap(@NonNull String projectKey, boolean ignoreCache);

    /**
     * @throws IllegalArgumentException if the repo key is not matching to the given projectKey prefix or the
     *                                  environments are not configured for the projectKey.
     */
    default void validateRepoConfigIsMatchingToProjectSpec(String projectKey, String repoKey,
            Set<String> environments) {
    }

    /**
     * @return The projectKey storage quota limit in bytes or {@link #UNLIMITED_STORAGE} if failed to find the
     * projectKey
     */
    long getProjectStorageQuotaBytes(String projectKey);

    /**
     * *** DO NOT USE IT ***
     * It does not verify authentication and super admin users
     * You should use {@link org.artifactory.api.security.AuthorizationService#isProjectAdmin}
     *
     * @param projectKey - Nullable for cases of default projectKey
     * @return True if the given user is an admin of the given projectKey
     */
    boolean isProjectAdmin(@Nullable String projectKey, @NonNull SimpleUser user);

    /**
     * *** DO NOT USE IT ***
     * It does not verify authentication and super admin users
     * You should use {@link org.artifactory.api.security.AuthorizationService#canManageProjectResources(String)} )}
     * or {@link org.artifactory.api.security.AuthorizationService#canManageResource(RepoPath)} )}
     *
     * @param projectKey - Nullable for cases of default projectKey
     * @return True if the given user can manage resources in the given projectKey
     */
    boolean canManageProjectResources(@Nullable String projectKey, @NonNull SimpleUser user);

    /**
     * *** DO NOT USE IT ***
     * It does not verify authentication and super admin users
     * You should use {@link org.artifactory.api.security.AuthorizationService#canManageResource(RepoPath)}
     *
     * @param repoKey - The repo key to check
     * @return True if the current user can manage the given repository
     */
    boolean canProjectAdminManageResources(String repoKey, @NonNull SimpleUser user);

    default void assertProjectStorageQuota(String repoKey, MutableStatusHolder statusHolder, long contentLength) {
    }

    /**
     * @return The repository's owned project storage quota limit in bytes. {@link ProjectsService#UNLIMITED_STORAGE}
     * serves as false for other cases
     */
    long getProjectQuotaSizeByRepository(@NonNull String repoKey);

    @NonNull
    Set<User> getProjectAdmins(@NonNull ProjectEntity projectKey);
}
